/*****************************************************************************
  bedFile.h

  (c) 2009 - Aaron Quinlan
  Hall Laboratory
  Department of Biochemistry and Molecular Genetics
  University of Virginia
  aaronquinlan@gmail.com

  Licensed under the GNU General Public License 2.0+ license.
******************************************************************************/
#ifndef BEDFILE_H
#define BEDFILE_H

#include "gzstream.h"
#include <vector>
#include <map>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cstring>
#include <algorithm>
#include <limits.h>
#include <stdint.h>
#include <cstdio>
//#include <tr1/unordered_map>  // Experimental.
using namespace std;


//*************************************************
// Data type tydedef
//*************************************************
typedef uint32_t CHRPOS;
typedef uint16_t BINLEVEL;
typedef uint32_t BIN;
typedef uint16_t USHORT;
typedef uint16_t UINT;

//*************************************************
// Genome binning constants
//*************************************************

const BIN      _numBins   = 37450;
const BINLEVEL _binLevels = 7;

// bins range in size from 16kb to 512Mb 
// Bin  0          spans 512Mbp,   # Level 1
// Bins 1-8        span 64Mbp,     # Level 2
// Bins 9-72       span 8Mbp,      # Level 3
// Bins 73-584     span 1Mbp       # Level 4
// Bins 585-4680   span 128Kbp     # Level 5
// Bins 4681-37449 span 16Kbp      # Level 6
const BIN _binOffsetsExtended[] = {32678+4096+512+64+8+1, 4096+512+64+8+1, 512+64+8+1, 64+8+1, 8+1, 1, 0};
	
const USHORT _binFirstShift = 14;	    /* How much to shift to get to finest bin. */
const USHORT _binNextShift  = 3;		/* How much to shift to get to next larger bin. */


//*************************************************
// Common data structures
//*************************************************

struct DEPTH {
	UINT starts;
	UINT ends;
};


/*
	Structure for regular BED records
*/
struct BED {

	// Regular BED fields
	string chrom;
	CHRPOS start;
	CHRPOS end; 
	string name;
	string score;
	string  strand;

	// Add'l fields for BED12 and/or custom BED annotations
	vector<string> otherFields;
	
public:
    // constructors

    // Null
    BED() 
    : chrom(""), 
      start(0),
      end(0),
      name(""),
      score(""),
      strand(""),
      otherFields()
    {}
        
    // BED3
    BED(string chrom, CHRPOS start, CHRPOS end) 
    : chrom(chrom), 
      start(start),
      end(end)
    {}

    // BED4
    BED(string chrom, CHRPOS start, CHRPOS end, string strand) 
    : chrom(chrom), 
      start(start),
      end(end),
      strand(strand)
    {}
}; // BED



/*
	Structure for regular BED COVERAGE records
*/
struct BEDCOV {

	// Regular BED fields
	CHRPOS start;
	CHRPOS end;
	
	string chrom;
	string name;
	string score;
	string strand;

	// Add'l fields for BED12 and/or custom BED annotations	
	vector<string> otherFields;

    // Additional fields specific to computing coverage
    map<unsigned int, DEPTH> depthMap;
    unsigned int count;
    CHRPOS minOverlapStart;
};


// enum to flag the state of a given line in a BED file.
enum BedLineStatus
{ 
    BED_INVALID = -1,
    BED_HEADER  = 0,
    BED_BLANK   = 1,
    BED_VALID   = 2
};


// return the genome "bin" for a feature with this start and end
inline
BIN getBin(CHRPOS start, CHRPOS end) {
    --end;
    start >>= _binFirstShift;
    end   >>= _binFirstShift;
    
    for (register short i = 0; i < _binLevels; ++i) {
        if (start == end) return _binOffsetsExtended[i] + start;
        start >>= _binNextShift;
        end   >>= _binNextShift;
    }
    cerr << "start " << start << ", end " << end << " out of range in findBin (max is 512M)" << endl;
    return 0;
}


// return the amount of overlap between two features.  Negative if none and the the 
// number of negative bases is the distance between the two.
inline 
int overlaps(CHRPOS aS, CHRPOS aE, CHRPOS bS, CHRPOS bE) {
	return min(aE, bE) - max(aS, bS);
}


// templated function to convert objects to strings
template <typename T>
std::string ToString(const T & value)
{
	std::stringstream ss;
	ss << value;
	return ss.str();
}


// Ancillary functions



// BED Sorting Methods 
bool sortByChrom(const BED &a, const BED &b);	
bool sortByStart(const BED &a, const BED &b);
bool sortBySizeAsc(const BED &a, const BED &b);
bool sortBySizeDesc(const BED &a, const BED &b);
bool sortByScoreAsc(const BED &a, const BED &b);
bool sortByScoreDesc(const BED &a, const BED &b);
bool byChromThenStart(BED const &a, BED const &b);



//*************************************************
// Data structure typedefs
//*************************************************
typedef vector<BED>    bedVector;
typedef vector<BEDCOV> bedCovVector;

typedef map<BIN, bedVector,    std::less<BIN> > binsToBeds;
typedef map<BIN, bedCovVector, std::less<BIN> > binsToBedCovs;

typedef map<string, binsToBeds, std::less<string> >    masterBedMap;
typedef map<string, binsToBedCovs, std::less<string> > masterBedCovMap;
typedef map<string, bedVector, std::less<string> >   masterBedMapNoBin;


// EXPERIMENTAL
//typedef tr1::unordered_map<int, vector<BED> > binsToBeds;
//typedef tr1::unordered_map<string, binsToBeds> masterBedMap;
//typedef map<string, bedVector, std::less<string> > masterBedMap;
//typedef vector<BED> masterBedMapNew[_numBins];
//typedef tr1::unordered_map<int, vector<BED> > binsToBedsNew;
//typedef tr1::unordered_map<string, binsToBeds> masterBedMapNew;




//************************************************
// BedFile Class methods and elements
//************************************************
class BedFile {

public:

	// Constructor 
	BedFile(string &);

	// Destructor
	~BedFile(void);
	
	// Open a BED file for reading (creates an istream pointer)
	void Open(void);
	
	// Close an opened BED file.
	void Close(void);
	
	// Get the next BED entry in an opened BED file.
	BedLineStatus GetNextBed (BED &bed, int &lineNum);

	// load a BED file into a map keyed by chrom, then bin. value is vector of BEDs
	void loadBedFileIntoMap();

	// load a BED file into a map keyed by chrom, then bin. value is vector of BEDCOVs
	void loadBedCovFileIntoMap();

	// load a BED file into a map keyed by chrom. value is vector of BEDs
	void loadBedFileIntoMapNoBin();	

	// Given a chrom, start, end and strand for a single feature,
	// search for all overlapping features in another BED file.
	// Searches through each relevant genome bin on the same chromosome
	// as the single feature. Note: Adapted from kent source "binKeeperFind"
	void FindOverlapsPerBin(string chrom, CHRPOS start, CHRPOS end, string strand, vector<BED> &hits, bool forceStrand);

	// return true if at least one overlap was found.  otherwise, return false.
	bool FindOneOrMoreOverlapsPerBin(string chrom, CHRPOS start, CHRPOS end, string strand, 
										bool forceStrand, float overlapFraction = 0.0);

	// return true if at least one __reciprocal__ overlap was found.  otherwise, return false.
	bool FindOneOrMoreReciprocalOverlapsPerBin(string chrom, CHRPOS start, CHRPOS end, string strand, 
													bool forceStrand, float overlapFraction = 0.0);
	
	// Given a chrom, start, end and strand for a single feature,
	// increment a the number of hits for each feature in B file
	// that the feature overlaps
	void countHits(const BED &a, bool forceStrand);
	
	// the bedfile with which this instance is associated
	string bedFile;
	unsigned int bedType;  // 3-6, 12 for BED
						   // 9 for GFF
	
	// Main data structires used by BEDTools
    masterBedCovMap   bedCovMap;
	masterBedMap      bedMap;
	masterBedMapNoBin bedMapNoBin;
						
private:
	
	// data
	bool _isGff;
	bool _isVcf;	
	istream *_bedStream;

    void setGff (bool isGff);
    void setVcf (bool isVcf);
    
    
    /******************************************************
    Private definitions to circumvent linker issues with
    templated member functions.
    *******************************************************/

    /*
        parseLine: converts a lineVector into either BED or BEDCOV (templated, hence in header to avoid linker issues.)
    */
    template <typename T>
    inline BedLineStatus parseLine (T &bed, const vector<string> &lineVector, int &lineNum) {

    	char *p2End, *p3End, *p4End, *p5End;
    	long l2, l3, l4, l5;

    	// bail out if we have a blank line
    	if (lineVector.size() == 0) 
    		return BED_BLANK;

    	if ((lineVector[0].find("track") == string::npos) && (lineVector[0].find("browser") == string::npos) && (lineVector[0].find("#") == string::npos) ) {

    		// we need at least 3 columns
    		if (lineVector.size() >= 3) {

    			// test if columns	2 and 3 are integers.  If so, assume BED.
    			l2 = strtol(lineVector[1].c_str(), &p2End, 10);
    			l3 = strtol(lineVector[2].c_str(), &p3End, 10);

    			// strtol  will set p2End or p3End to the start of the string if non-integral, base 10
    			if (p2End != lineVector[1].c_str() && p3End != lineVector[2].c_str()) {
    				setGff(false);
    				if (parseBedLine(bed, lineVector, lineNum) == true)
    					return BED_VALID;
    			}
    			else if (p2End != lineVector[1].c_str()) {
                    setGff(false);
                    setVcf(true);
                    if (parseVcfLine(bed, lineVector, lineNum) == true)
                        return BED_VALID;
    			}
    			else if (lineVector.size() == 9) {
    				// otherwise test if columns 4 and 5 are integers.  If so, assume GFF.
    				l4 = strtol(lineVector[3].c_str(), &p4End, 10);
    				l5 = strtol(lineVector[4].c_str(), &p5End, 10);

    				// strtol  will set p4End or p5End to the start of the string if non-integral, base 10
    				if (p4End != lineVector[3].c_str() && p5End != lineVector[4].c_str()) {		
    					setGff(true);
    					if (parseGffLine(bed, lineVector, lineNum) == true)
    						return BED_VALID;
    				}
    			}
    			else {
    				cerr << "Unexpected file format.  Please use tab-delimited BED or GFF" << endl;
    				exit(1);
    			}
    		}
    		// gripe if it's not a blank line	
    		else if (lineVector.size() != 0) {
    			cerr << "It looks as though you have less than 3 columns at line: " << lineNum << ".  Are you sure your files are tab-delimited?" << endl;
    			exit(1);
    		}
    	}
    	else {
    		lineNum--;
    		return BED_HEADER;	
    	}

    	// default
    	return BED_INVALID;
    }


    /*
        parseBedLine: converts a lineVector into either BED or BEDCOV (templated, hence in header to avoid linker issues.)
    */
    template <typename T>
    inline bool parseBedLine (T &bed, const vector<string> &lineVector, int lineNum) {

    	if ( (lineNum > 1) && (lineVector.size() == this->bedType)) {

    		bed.chrom = lineVector[0];
    		bed.start = atoi(lineVector[1].c_str());
    		bed.end = atoi(lineVector[2].c_str());

    		if (this->bedType == 4) {
    			bed.name = lineVector[3];
    		}
    		else if (this->bedType ==5) {
    			bed.name = lineVector[3];
    			bed.score = lineVector[4];
    		}
    		else if (this->bedType == 6) {
    			bed.name = lineVector[3];
    			bed.score = lineVector[4];
    			bed.strand = lineVector[5];
    		}
    		else if (this->bedType > 6) {
    			bed.name = lineVector[3];
    			bed.score = lineVector[4];
    			bed.strand = lineVector[5];			
    			for (unsigned int i = 6; i < lineVector.size(); ++i) {
    				bed.otherFields.push_back(lineVector[i]); 
    			}
    		}
    		else if (this->bedType != 3) {
    			cerr << "Error: unexpected number of fields at line: " << lineNum << ".  Verify that your files are TAB-delimited and that your BED file has 3,4,5 or 6 fields.  Exiting..." << endl;
    			exit(1);
    		}

    		if ((bed.start <= bed.end) && (bed.start > 0) && (bed.end > 0)) {
                return true;
    		}
    		else if (bed.start > bed.end) {
    			cerr << "Error: malformed BED entry at line " << lineNum << ". Start was greater than end. Exiting." << endl;
    			exit(1);
    		}
    		else if ( (bed.start < 0) || (bed.end < 0) ) {
    			cerr << "Error: malformed BED entry at line " << lineNum << ". Coordinate detected that is < 0. Exiting." << endl;
    			exit(1);
    		}
    	}
    	else if ((lineNum == 1) && (lineVector.size() >= 3)) {
    		this->bedType = lineVector.size();

    		bed.chrom = lineVector[0];
    		bed.start = atoi(lineVector[1].c_str());
    		bed.end = atoi(lineVector[2].c_str());

            if (this->bedType == 4) {
    			bed.name = lineVector[3];
    		}
    		else if (this->bedType ==5) {
    			bed.name = lineVector[3];
    			bed.score = lineVector[4];
    		}
    		else if (this->bedType == 6) {
    			bed.name = lineVector[3];
    			bed.score = lineVector[4];
    			bed.strand = lineVector[5];
    		}
    		else if (this->bedType > 6) {
    			bed.name = lineVector[3];
    			bed.score = lineVector[4];
    			bed.strand = lineVector[5];	
    			for (unsigned int i = 6; i < lineVector.size(); ++i) {
    				bed.otherFields.push_back(lineVector[i]); 
    			}
    		}
    		else if (this->bedType != 3) {
    			cerr << "Error: unexpected number of fields at line: " << lineNum << ".  Verify that your files are TAB-delimited and that your BED file has 3,4,5 or 6 fields.  Exiting..." << endl;
    			exit(1);
    		}
    		if ((bed.start <= bed.end) && (bed.start > 0) && (bed.end > 0)) {
                return true;
    		}
    		else if (bed.start > bed.end) {
    			cerr << "Error: malformed BED entry at line " << lineNum << ". Start was greater than end. Exiting." << endl;
    			exit(1);
    		}
    		else if ( (bed.start < 0) || (bed.end < 0) ) {
    			cerr << "Error: malformed BED entry at line " << lineNum << ". Coordinate detected that is < 0. Exiting." << endl;
    			exit(1);
    		}
    	}
    	else if (lineVector.size() == 1) {
    		cerr << "Only one BED field detected: " << lineNum << ".  Verify that your files are TAB-delimited.  Exiting..." << endl;
    		exit(1);		
    	}
    	else if ((lineVector.size() != this->bedType) && (lineVector.size() != 0)) {
    		cerr << "Differing number of BED fields encountered at line: " << lineNum << ".  Exiting..." << endl;
    		exit(1);
    	}
    	else if ((lineVector.size() < 3) && (lineVector.size() != 0)) {
    		cerr << "TAB delimited BED file with at least 3 fields (chrom, start, end) is required at line: "<< lineNum << ".  Exiting..." << endl;
    		exit(1);
    	}
    	return false;
    }


    /*
        parseBedLine: converts a lineVector into either BED or BEDCOV (templated, hence in header to avoid linker issues.)
    */
    template <typename T>
    inline bool parseVcfLine (T &bed, const vector<string> &lineVector, int lineNum) {

    	if ( (lineNum > 1) && (lineVector.size() == this->bedType)) {

    		bed.chrom  = lineVector[0];
    		bed.start  = atoi(lineVector[1].c_str()) - 1;  // VCF is one-based
    		bed.end    = atoi(lineVector[1].c_str());      // TO-DO: make this the size of the variant.
            bed.strand = "+";

    		if (this->bedType > 2) {		
    			for (unsigned int i = 2; i < lineVector.size(); ++i) {
    				bed.otherFields.push_back(lineVector[i]); 
    			}
    		}

    		if ((bed.start <= bed.end) && (bed.start > 0) && (bed.end > 0)) {
                return true;
    		}
    		else if (bed.start > bed.end) {
    			cerr << "Error: malformed VCF entry at line " << lineNum << ". Start was greater than end. Exiting." << endl;
    			exit(1);
    		}
    		else if ( (bed.start < 0) || (bed.end < 0) ) {
    			cerr << "Error: malformed VCF entry at line " << lineNum << ". Coordinate detected that is < 0. Exiting." << endl;
    			exit(1);
    		}
    	}
    	else if ((lineNum == 1) && (lineVector.size() >= 6)) {
    		this->bedType = lineVector.size();

    		bed.chrom  = lineVector[0];
    		bed.start  = atoi(lineVector[1].c_str()) - 1;  // VCF is one-based
    		bed.end    = atoi(lineVector[1].c_str());      // TO-DO: make this the size of the variant.
            bed.strand = "+";

    		if (this->bedType > 2) {		
    			for (unsigned int i = 2; i < lineVector.size(); ++i) {
    				bed.otherFields.push_back(lineVector[i]); 
    			}
    		}

    		if ((bed.start <= bed.end) && (bed.start > 0) && (bed.end > 0)) {
                return true;
    		}
    		else if (bed.start > bed.end) {
    			cerr << "Error: malformed VCF entry at line " << lineNum << ". Start was greater than end. Exiting." << endl;
    			exit(1);
    		}
    		else if ( (bed.start < 0) || (bed.end < 0) ) {
    			cerr << "Error: malformed VCF entry at line " << lineNum << ". Coordinate detected that is < 0. Exiting." << endl;
    			exit(1);
    		}
    	}
    	else if (lineVector.size() == 1) {
    		cerr << "Only one VCF field detected: " << lineNum << ".  Verify that your files are TAB-delimited.  Exiting..." << endl;
    		exit(1);		
    	}
    	else if ((lineVector.size() != this->bedType) && (lineVector.size() != 0)) {
    		cerr << "Differing number of VCF fields encountered at line: " << lineNum << ".  Exiting..." << endl;
    		exit(1);
    	}
    	else if ((lineVector.size() < 2) && (lineVector.size() != 0)) {
    		cerr << "TAB delimited VCF file with at least 2 fields (chrom, pos) is required at line: "<< lineNum << ".  Exiting..." << endl;
            exit(1);
    	}
    	return false;
    }



    /*
        parseGffLine: converts a lineVector into either BED or BEDCOV (templated, hence in header to avoid linker issues.)
    */
    template <typename T>
    bool parseGffLine (T &bed, const vector<string> &lineVector, int lineNum) {

    /*
    1.  seqname - The name of the sequence. Must be a chromosome or scaffold.
    2. source - The program that generated this feature.
    3. feature - The name of this type of feature. Some examples of standard feature types are "CDS", 
    			"start_codon", "stop_codon", and "exon".
    4. start - The starting position of the feature in the sequence. The first base is numbered 1.
    5. end - The ending position of the feature (inclusive).
    6. score - A score between 0 and 1000. If the track line useScore attribute is set to 1 
    			for this annotation data set, the score value will determine the level of gray 
    			in which this feature is displayed (higher numbers = darker gray). 
    			If there is no score value, enter ".".
    7. strand - Valid entries include '+', '-', or '.' (for don't know/don't care).
    8. frame - If the feature is a coding exon, frame should be a number between 0-2 that 
    			represents the reading frame of the first base. If the feature is not a coding exon, 
    			the value should be '.'.
    9. group - All lines with the same group are linked together into a single item.
    */
    	if ( (lineNum > 1) && (lineVector.size() == this->bedType)) {

    		if (this->bedType == 9 && _isGff) {
    			bed.chrom = lineVector[0];
    			// substract 1 to force the start to be BED-style
    			bed.start = atoi(lineVector[3].c_str()) - 1;
    			bed.end = atoi(lineVector[4].c_str());
    			bed.name = lineVector[2];
    			bed.score = lineVector[5];
    			bed.strand = atoi(lineVector[6].c_str());
    			bed.otherFields.push_back(lineVector[1]);  // add GFF "source". unused in BED
    			bed.otherFields.push_back(lineVector[7]);  // add GFF "fname". unused in BED
    			bed.otherFields.push_back(lineVector[8]);  // add GFF "group". unused in BED
    		}
    		else {
    			cerr << "Error: unexpected number of fields at line: " << lineNum << 
    					".  Verify that your files are TAB-delimited and that your GFF file has 9 fields.  Exiting..." << endl;
    			exit(1);
    		}

    		if (bed.start > bed.end) {
    			cerr << "Error: malformed GFF entry at line " << lineNum << ". Start was greater than end. Exiting." << endl;
    			exit(1);
    		}
    		else if ( (bed.start < 0) || (bed.end < 0) ) {
    			cerr << "Error: malformed GFF entry at line " << lineNum << ". Coordinate detected that is < 1. Exiting." << endl;
    			exit(1);
    		}
    		else return true;
    	}
    	else if ((lineNum == 1) && (lineVector.size() == 9)) {
    		this->bedType = lineVector.size();

    		if (this->bedType == 9 && _isGff) {
    			bed.chrom = lineVector[0];
    			// substract 1 to force the start to be BED-style
    			bed.start = atoi(lineVector[3].c_str()) - 1;
    			bed.end = atoi(lineVector[4].c_str());
    			bed.name = lineVector[2];
    			bed.score = lineVector[5];
    			bed.strand = atoi(lineVector[6].c_str());

    			bed.otherFields.push_back(lineVector[1]);  // add GFF "source". unused in BED
    			bed.otherFields.push_back(lineVector[7]);  // add GFF "fname".  unused in BED
    			bed.otherFields.push_back(lineVector[8]);  // add GFF "group".  unused in BED

    			return true;
    		}
    		else {
    			cout << this->bedType << " " << _isGff << endl;
    			cerr << "Error: unexpected number of fields at line: " << lineNum << 
    					".  Verify that your files are TAB-delimited and that your GFF file has 9 fields.  Exiting..." << endl;
    			exit(1);
    		}

    		if (bed.start > bed.end) {
    			cerr << "Error: malformed GFF entry at line " << lineNum << ". Start was greater than end. Exiting." << endl;
    			exit(1);
    		}
    		else if ( (bed.start < 0) || (bed.end < 0) ) {
    			cerr << "Error: malformed GFF entry at line " << lineNum << ". Coordinate detected that is < 1. Exiting." << endl;
    			exit(1);
    		}
    		else return true;
    	}
    	else if (lineVector.size() == 1) {
    		cerr << "Only one GFF field detected: " << lineNum << ".  Verify that your files are TAB-delimited.  Exiting..." << endl;
    		exit(1);		
    	}
    	else if ((lineVector.size() != this->bedType) && (lineVector.size() != 0)) {
    		cerr << "Differing number of GFF fields encountered at line: " << lineNum << ".  Exiting..." << endl;
    		exit(1);
    	}
    	else if ((lineVector.size() < 9) && (lineVector.size() != 0)) {
    		cerr << "TAB delimited GFF file with 9 fields is required at line: "<< lineNum << ".  Exiting..." << endl;
    		exit(1);
    	}
    	return false;
    }
    
	
public:
    
    /*
    	reportBedTab

    	Writes the _original_ BED entry with a TAB
    	at the end of the line.
    	Works for BED3 - BED6.
    */
    template <typename T>
    void reportBedTab(const T &bed) {
    	if (_isGff == false) {
    		if (this->bedType == 3) {
    			printf ("%s\t%d\t%d\t", bed.chrom.c_str(), bed.start, bed.end);
    		}
    		else if (this->bedType == 4) {
    			printf ("%s\t%d\t%d\t%s\t", bed.chrom.c_str(), bed.start, bed.end, bed.name.c_str());
    		}
    		else if (this->bedType == 5) {
    			printf ("%s\t%d\t%d\t%s\t%s\t", bed.chrom.c_str(), bed.start, bed.end, bed.name.c_str(), 
    											bed.score.c_str());
    		}
    		else if (this->bedType == 6) {
    			printf ("%s\t%d\t%d\t%s\t%s\t%s\t", bed.chrom.c_str(), bed.start, bed.end, bed.name.c_str(), 
    												bed.score.c_str(), bed.strand.c_str());
    		}
    		else if (this->bedType > 6) {
    			printf ("%s\t%d\t%d\t%s\t%s\t%s\t", bed.chrom.c_str(), bed.start, bed.end, bed.name.c_str(), 
    												bed.score.c_str(), bed.strand.c_str());

    			vector<string>::const_iterator othIt = bed.otherFields.begin(); 
    			vector<string>::const_iterator othEnd = bed.otherFields.end(); 
    			for ( ; othIt != othEnd; ++othIt) {
    				printf("%s\t", othIt->c_str());
    			}
    		}
    	}	
    	else if (this->bedType == 9) {
    		printf ("%s\t%s\t%s\t%d\t%d\t%s\t%s\t%s\t%s\t", bed.chrom.c_str(), bed.otherFields[0].c_str(),
    														 bed.name.c_str(), bed.start+1, bed.end, 
    														 bed.score.c_str(), bed.strand.c_str(),
    														 bed.otherFields[1].c_str(), bed.otherFields[2].c_str());
    	}
    }



    /*
    	reportBedNewLine

    	Writes the _original_ BED entry with a NEWLINE
    	at the end of the line.
    	Works for BED3 - BED6.
    */
    template <typename T>
    void reportBedNewLine(const T &bed) {
    	if (_isGff == false) {
    		if (this->bedType == 3) {
    			printf ("%s\t%d\t%d\n", bed.chrom.c_str(), bed.start, bed.end);
    		}
    		else if (this->bedType == 4) {
    			printf ("%s\t%d\t%d\t%s\n", bed.chrom.c_str(), bed.start, bed.end, bed.name.c_str());
    		}
    		else if (this->bedType == 5) {
    			printf ("%s\t%d\t%d\t%s\t%s\n", bed.chrom.c_str(), bed.start, bed.end, bed.name.c_str(), 
    											bed.score.c_str());
    		}
    		else if (this->bedType == 6) {
    			printf ("%s\t%d\t%d\t%s\t%s\t%s\n", bed.chrom.c_str(), bed.start, bed.end, bed.name.c_str(), 
    												bed.score.c_str(), bed.strand.c_str());
    		}
    		else if (this->bedType > 6) {
    			printf ("%s\t%d\t%d\t%s\t%s\t%s\t", bed.chrom.c_str(), bed.start, bed.end, bed.name.c_str(), 
    												bed.score.c_str(), bed.strand.c_str());

    			vector<string>::const_iterator othIt = bed.otherFields.begin(); 
    			vector<string>::const_iterator othEnd = bed.otherFields.end(); 
    			for ( ; othIt != othEnd; ++othIt) {
    				printf("%s\t", othIt->c_str());
    			}
    			printf("\n");
    		}
    	}
    	else if (this->bedType == 9) {
    		printf ("%s\t%s\t%s\t%d\t%d\t%s\t%s\t%s\t%s\n", bed.chrom.c_str(), bed.otherFields[0].c_str(),
    														 bed.name.c_str(), bed.start+1, bed.end, 
    														 bed.score.c_str(), bed.strand.c_str(),
    														 bed.otherFields[1].c_str(), bed.otherFields[2].c_str());
    	}
    }



    /*
    	reportBedRangeNewLine

    	Writes a custom start->end for a BED entry
    	with a NEWLINE at the end of the line.

    	Works for BED3 - BED6.
    */
    template <typename T>
    void reportBedRangeTab(const T &bed, CHRPOS start, CHRPOS end) {
    	if (_isGff == false) {
    		if (this->bedType == 3) {
    			printf ("%s\t%d\t%d\t", bed.chrom.c_str(), start, end);
    		}
    		else if (this->bedType == 4) {
    			printf ("%s\t%d\t%d\t%s\t", bed.chrom.c_str(), start, end, bed.name.c_str());
    		}
    		else if (this->bedType == 5) {
    			printf ("%s\t%d\t%d\t%s\t%s\t", bed.chrom.c_str(), start, end, bed.name.c_str(), 
    											bed.score.c_str());
    		}
    		else if (this->bedType == 6) {
    			printf ("%s\t%d\t%d\t%s\t%s\t%s\t", bed.chrom.c_str(), start, end, bed.name.c_str(), 
    												bed.score.c_str(), bed.strand.c_str());
    		}
    		else if (this->bedType > 6) {
    			printf ("%s\t%d\t%d\t%s\t%s\t%s\t", bed.chrom.c_str(), start, end, bed.name.c_str(), 
    												bed.score.c_str(), bed.strand.c_str());

    			vector<string>::const_iterator othIt = bed.otherFields.begin(); 
    			vector<string>::const_iterator othEnd = bed.otherFields.end(); 
    			for ( ; othIt != othEnd; ++othIt) {
    				printf("%s\t", othIt->c_str());
    			}
    		}
    	}
    	else if (this->bedType == 9) {
    		printf ("%s\t%s\t%s\t%d\t%d\t%s\t%s\t%s\t%s\t", bed.chrom.c_str(), bed.otherFields[0].c_str(),
    														 bed.name.c_str(), start+1, end, 
    														 bed.score.c_str(), bed.strand.c_str(),
    														 bed.otherFields[1].c_str(), bed.otherFields[2].c_str());
    	}
    }



    /*
    	reportBedRangeTab

    	Writes a custom start->end for a BED entry 
    	with a TAB at the end of the line.

    	Works for BED3 - BED6.
    */
    template <typename T>
    void reportBedRangeNewLine(const T &bed, CHRPOS start, CHRPOS end) {
    	if (_isGff == false) {
    		if (this->bedType == 3) {
    			printf ("%s\t%d\t%d\n", bed.chrom.c_str(), start, end);
    		}
    		else if (this->bedType == 4) {
    			printf ("%s\t%d\t%d\t%s\n", bed.chrom.c_str(), start, end, bed.name.c_str());
    		}
    		else if (this->bedType == 5) {
    			printf ("%s\t%d\t%d\t%s\t%s\n", bed.chrom.c_str(), start, end, bed.name.c_str(), 
    											bed.score.c_str());
    		}
    		else if (this->bedType == 6) {
    			printf ("%s\t%d\t%d\t%s\t%s\t%s\n", bed.chrom.c_str(), start, end, bed.name.c_str(), 
    												bed.score.c_str(), bed.strand.c_str());
    		}
    		else if (this->bedType > 6) {
    			printf ("%s\t%d\t%d\t%s\t%s\t%s\t", bed.chrom.c_str(), start, end, bed.name.c_str(), 
    												bed.score.c_str(), bed.strand.c_str());

    			vector<string>::const_iterator othIt = bed.otherFields.begin(); 
    			vector<string>::const_iterator othEnd = bed.otherFields.end(); 
    			for ( ; othIt != othEnd; ++othIt) {
    				printf("%s\t", othIt->c_str());
    			}
    			printf("\n");
    		}
    	}
    	else if (this->bedType == 9) {	// add 1 to the start for GFF
    		printf ("%s\t%s\t%s\t%d\t%d\t%s\t%s\t%s\t%s\n", bed.chrom.c_str(), bed.otherFields[0].c_str(),
    														 bed.name.c_str(), start+1, end, 
    														 bed.score.c_str(), bed.strand.c_str(),
    														 bed.otherFields[1].c_str(), bed.otherFields[2].c_str());
    	}
    }


    /*
    	reportNullBedTab
    */
    void reportNullBedTab() {

    	if (_isGff == false) {
    		if (this->bedType == 3) {
    			printf (".\t-1\t-1\t");
    		}
    		else if (this->bedType == 4) {
    			printf (".\t-1\t-1\t.\t");
    		}
    		else if (this->bedType == 5) {
    			printf (".\t-1\t-1\t.\t-1\t");
    		}
    		else if (this->bedType == 6) {
    			printf (".\t-1\t-1\t.\t-1\t.\t");
    		}
    		else if (this->bedType > 6) {
    			printf (".\t-1\t-1\t.\t-1\t.\t");
    			for (unsigned int i = 6; i < this->bedType; ++i) {
    				printf(".\t");
    			}
    		}
    	}	
    	else if (this->bedType == 9) {
    		printf (".\t.\t.\t-1\t-1\t-1\t.\t.\t.\t");
    	}
    }


    /*
    	reportNullBedTab
    */
    void reportNullBedNewLine() {

    	if (_isGff == false) {
    		if (this->bedType == 3) {
    			printf (".\t-1\t-1\n");
    		}
    		else if (this->bedType == 4) {
    			printf (".\t-1\t-1\t.\n");
    		}
    		else if (this->bedType == 5) {
    			printf (".\t-1\t-1\t.\t-1\n");
    		}
    		else if (this->bedType == 6) {
    			printf (".\t-1\t-1\t.\t-1\t.\n");
    		}
    		else if (this->bedType > 6) {
    			printf (".\t-1\t-1\t.\t-1\t.\n");
    			for (unsigned int i = 6; i < this->bedType; ++i) {
    				printf(".\t");
    			}
    			printf("\n");
    		}
    	}	
    	else if (this->bedType == 9) {
    		printf (".\t.\t.\t-1\t-1\t-1\t.\t.\t.\n");
    	}
    }

    
};

#endif /* BEDFILE_H */
